module Assembler where

import Char

import Lexer
import Instruction
import Table

assemble c =
    resolvAddress is table
    where
      result = assembleInstructions 0 (initialTable (\x -> -1)) c
      is = fst result
      table = snd result

assembleInstructions address table ((ins, comment):cs) = 
    ((address, machineCodes, ins, comment) : fst next, snd next)
    where
      machineCodes = assembleInstruction address ins
      machineCodesLength = length machineCodes
      newTable = registerLabel table ins address
      next = assembleInstructions (address + machineCodesLength) newTable cs
assembleInstructions address table [] = ([], table)

registerLabel table (Label l) address = updateTable table l address
registerLabel table _ address = table

resolvAddress ((address, machineCodes, ins, comment):is) table =
    case ins of
      (OperandString1 "JUMP" l) ->
          let
              tableAddress = contentTable table l
              labelAddress = if tableAddress >= 0 then tableAddress
                             else if (length l) == 2 &&
                                      (any (toUpper (head l)==) "ABCDEF" || isDigit (head l))
                                   then readHexString l
                                   else error ("assemble error: undeclared label " ++ l)

              labelAddressCode = [div labelAddress 16, mod labelAddress 16]
          in
            (address, [0xf] ++ labelAddressCode, ins, comment)
      _ -> (address, machineCodes, ins, comment)
    : resolvAddress is table
resolvAddress [] table = []


assembleError = error . ("assemble error: " ++)

assembleInstruction _ (Operand0 "KA") = [0x0]
assembleInstruction _ (Operand0 "AO") = [0x1]
assembleInstruction _ (Operand0 "CH") = [0x2]
assembleInstruction _ (Operand0 "CY") = [0x3]
assembleInstruction _ (Operand0 "AM") = [0x4]
assembleInstruction _ (Operand0 "MA") = [0x5]
assembleInstruction _ (Operand0 "M+") = [0x6]
assembleInstruction _ (Operand0 "M-") = [0x7]
assembleInstruction _ (OperandInt1 "TIA" op1) = [0x8, op1]
assembleInstruction _ (OperandInt1 "AIA" op1) = [0x9, op1]
assembleInstruction _ (OperandInt1 "TIY" op1) = [0xa, op1]
assembleInstruction _ (OperandInt1 "AIY" op1) = [0xb, op1]
assembleInstruction _ (OperandInt1 "CIA" op1) = [0xc, op1]
assembleInstruction _ (OperandInt1 "CIY" op1) = [0xd, op1]
assembleInstruction _ (OperandString1 "JUMP" _) = [0xf, 0, 0]
assembleInstruction _ (OperandString1 "CAL" "RSTO") = [0xe, 0x0]
assembleInstruction _ (OperandString1 "CAL" "SETR") = [0xe, 0x1]
assembleInstruction _ (OperandString1 "CAL" "RSTR") = [0xe, 0x2]
assembleInstruction _ (OperandString1 "CAL" "INPT") = [0xe, 0x3]
assembleInstruction _ (OperandString1 "CAL" "CMPL") = [0xe, 0x4]
assembleInstruction _ (OperandString1 "CAL" "CHNG") = [0xe, 0x5]
assembleInstruction _ (OperandString1 "CAL" "SIFT") = [0xe, 0x6]
assembleInstruction _ (OperandString1 "CAL" "ENDS") = [0xe, 0x7]
assembleInstruction _ (OperandString1 "CAL" "ERRS") = [0xe, 0x8]
assembleInstruction _ (OperandString1 "CAL" "SHTS") = [0xe, 0x9]
assembleInstruction _ (OperandString1 "CAL" "LONS") = [0xe, 0xa]
assembleInstruction _ (OperandString1 "CAL" "SUND") = [0xe, 0xb]
assembleInstruction _ (OperandString1 "CAL" "TIMR") = [0xe, 0xc]
assembleInstruction _ (OperandString1 "CAL" "DSPR") = [0xe, 0xd]
assembleInstruction _ (OperandString1 "CAL" "DEM-") = [0xe, 0xe]
assembleInstruction _ (OperandString1 "CAL" "DEM+") = [0xe, 0xf]
assembleInstruction _ (Label _) = []

assembleInstruction _ (OperandInt1 ".NIBBLE" op1) = [op1]
assembleInstruction address (OperandString1 ".ALIGN" num) = 
    if align == 0
    then []
    else take remain $ repeat 0xf
    where
      align = readHexString num
      remain = align - (mod address align)


assembleInstruction _ x = assembleError ("unkown instruction " ++ show x)

