module Emitter where

import List

import Lexer
import Parser
import Table
import Instruction

data Environment = Environment (Table (Type, Int), Int, Int) deriving Show

envTable (Environment (table, offset, label)) = table
envOffset (Environment (table, offset, label)) = offset
envLabel (Environment (table, offset, label)) = label
incEnvOffset (Environment (table, offset, label)) = 
    (Environment (table, offset - 1, label))
decEnvOffset (Environment (table, offset, label)) = 
    (Environment (table, offset + 1, label))
incEnvLabel (Environment (table, offset, label)) = (Environment (table, offset , label + 1))

data EmitValue =
    EmitValueVoid |
    EmitValueIntRegisterA |
    EmitValueIntRegisterY |
    EmitValueIntTemporaryMemory Int |
    EmitValueIntVariable Int String |
    EmitValueIntConstant Int |
    EmitValueAddressRegisterA |
    EmitValueAddressRegisterY |
    EmitValueAddress Int String |
    EmitValueAddressInMemory Int String |
    EmitValueLabelsFalse [String] |
    EmitValueLabelsTrue [String] |
    EmitValueNone deriving (Show)

data EmitResult =
    EmitResult (EmitValue, [(Instruction, String)], Environment) |
    EmitResultAlternative EmitResult [EmitResult]
    deriving Show

emitResultValue (EmitResult (ev, code, env)) = ev
emitResultValue (EmitResultAlternative er _) = emitResultValue er
emitResultCode (EmitResult (ev, code, env)) = code
emitResultCode (EmitResultAlternative er _) = emitResultCode er
emitResultEnvironment (EmitResult (ev, code, env)) = env
emitResultEnvironment (EmitResultAlternative er _) = emitResultEnvironment er

emitResultSelect p er@(EmitResult _) = er
emitResultSelect p (EmitResultAlternative der (er:ers)) | p er = er
emitResultSelect p (EmitResultAlternative der (er:ers)) = emitResultSelect p (EmitResultAlternative der ers)
emitResultSelect p (EmitResultAlternative der []) = der

emitResultIsFalseLabels (EmitResult (EmitValueLabelsFalse _, _, _)) = True
emitResultIsFalseLabels _ = False

emitResultIsTrueLabels (EmitResult (EmitValueLabelsTrue _, _, _)) = True
emitResultIsTrueLabels _ = False

data EmitContext = EmitContext { emitContextBreakLabel::String, emitContextNeedValue::Bool }

emitError = error . ("emitter error: " ++)

emit tree = emitProgram tree

emitProgram tree@(Program globals) = 
    functionCodes ++
    if declarations == []
    then []
    else 
        [(OperandString1 ".ALIGN" (hexString 0 (0x50 + imageOffset)), "Global variables")] ++ image
    where
      setLedBinaryArguments = getSetLedBinaryArgumentsProgram tree
      initialFixedTable = initialTable (\x -> -1)
      fixedTable =
          case setLedBinaryArguments of
            [(Operation2 (",",VariableValue a,VariableValue b))] ->
                updateTable (updateTable initialFixedTable a 0xf) b 0xe
            _ -> initialFixedTable

      env =
          (Environment (initialTable (\x -> error ("emmiter error: undeclared variable " ++ show x)), 
                        if setLedBinaryArguments == [] then 0xf else 0xd, 0))

      declarations = fst $ splitGlobals globals
      newEnv = emitDeclarationsWithFixedOffsetTable fixedTable declarations env
      functions = snd $ splitGlobals globals
      functionCodes = emitConcat emitFunction newEnv functions

      rowImage = emitGlobalInitializations declarations newEnv initialImage
      imageOffset = fst (cutImage rowImage)
      image = map recoveryDummyInstruction (snd (cutImage rowImage))
      dummyInstruction = (OperandInt1 ".NIBBLE_DUMMY" 0xf, "")
      initialImage = take 0x10 $ repeat dummyInstruction
      cutImage = cutImage' 0
      cutImage' n (i:is) | i == dummyInstruction = cutImage' (n + 1) is
      cutImage' n is = (n, is)
      recoveryDummyInstruction i | i == dummyInstruction = (OperandInt1 ".NIBBLE" 0xf, "")
      recoveryDummyInstruction i = i

splitGlobals (g:gs) =
    case g of
      (GlobalDeclaration x) ->
          (x:(fst next),snd next)
      (GlobalFunction x) ->
          ((fst next),x: snd next)
    where
      next = splitGlobals gs
splitGlobals [] = ([], [])

emitGlobalInitializations declarations env image = 
    foldl (\x y -> emitGlobalInitialization' y env x) image declarations
emitGlobalInitialization' (Declaration variables) env image =
    foldl (\x y -> emitGlobalInitialization y env x) image variables
emitGlobalInitialization (Variable (TypeInt, vname, 1), InitializationInt i) env image = 
    overwriteDataImage (snd (contentTable (envTable env) vname)) i ("Global variable '" ++ vname ++ "'") image
emitGlobalInitialization (Variable (TypeInt, vname, 1), InitializationNone) env image = 
    overwriteDataImage (snd (contentTable (envTable env) vname)) 0 ("Global variable '" ++ vname ++ "'") image
emitGlobalInitialization (Variable (TypePointer, vname, 1), InitializationNone) env image = 
    overwriteDataImage (snd (contentTable (envTable env) vname)) 0 ("Global variable '" ++ vname ++ "'") image
emitGlobalInitialization (Variable (TypeArray, vname, num), InitializationIntArray is) env image = 
    emitGlobalInitializationIntArray 0 vname num is env image
emitGlobalInitialization (_, InitializationNone) env image = image
emitGlobalInitialization _ env image = emitError "invalid global initialization"

emitGlobalInitializationIntArray index vname num (i:is) env image
    | index < num || num == 0 =
        emitGlobalInitializationIntArray (index + 1) vname num is env newImage
    where
      newImage = overwriteDataImage ((snd (contentTable (envTable env) vname)) + index) i 
                 ("Global array '" ++ vname ++ "[" ++ show index ++ "]'") image
emitGlobalInitializationIntArray index vname num [] env image = image
emitGlobalInitializationIntArray _ _ _ _ _ _ = emitError "invalid global array initialization"


overwriteDataImage 0 d comment (i:is) = (OperandInt1 ".NIBBLE" (immediateValue d), comment):is
overwriteDataImage n d comment (i:is) = i:overwriteDataImage (n - 1) d comment is

emitConcat f env l = concat $ map (\x -> f x env) l

emitFunction (Function (name, ftype, arguments, block)) env =
    emitResultCode (emitBlock block env (EmitContext {emitContextBreakLabel = "", emitContextNeedValue = False}))

emitBlock (Block (declarations, statements)) env context =
    EmitResult(emitResultValue statementsResult,
               emitResultCode statementsResult,
               Environment (envTable env, envOffset env, envLabel (emitResultEnvironment statementsResult)))
    where
      newEnv = emitDeclarations declarations env
      initializationStatements = emitInitializations declarations env
      statementsResult = emitStatements (initializationStatements ++ statements) newEnv context
      
emitDeclarations = emitDeclarationsWithFixedOffsetTable (initialTable (\x -> -1))

emitDeclarationsWithFixedOffsetTable fixedTable declarations env =
    foldl (\x y -> emitDeclaration fixedTable y x) env declarations

emitDeclaration fixedTable (Declaration variables) env =
    foldl (\x y -> emitDeclarationVariable fixedTable y x) env variables

emitDeclarationVariable fixedTable (Variable (TypeArray, vname, num), (InitializationIntArray is))
    | num == 0 = emitDeclarationVariableAllocate fixedTable TypeArray vname (length is)
emitDeclarationVariable fixedTable (Variable (TypeArray, vname, num), _)
    | num > 0 = emitDeclarationVariableAllocate fixedTable TypeArray vname num
emitDeclarationVariable fixedTable (Variable (TypeArray, vname, num), _)
    | num < 0 = emitError "invalid array size"
emitDeclarationVariable fixedTable (Variable (vtype, vname, 1), _) =
    emitDeclarationVariableAllocate fixedTable vtype vname 1
emitDeclarationVariable fixedTable _ = emitError "invalid declaration"

emitDeclarationVariableAllocate fixedTable vtype vname vsize env
    | contentTable fixedTable vname >= 0 =
        Environment (newTable, envOffset env, envLabel env)
    where
      newTable = updateTable (envTable env) vname (vtype, contentTable fixedTable vname)

emitDeclarationVariableAllocate fixedTable vtype vname vsize env =
    if offset - vsize < -1 then
        emitError "memory over"
    else
        Environment (newTable, offset - vsize, envLabel env)
    where
      offset = envOffset env
      newTable = updateTable (envTable env) vname (vtype, offset - vsize + 1)

emitInitializations declarations env =
    concat $ map (\x -> emitInitialization x) (concat (map (\(Declaration x) -> x) declarations))
emitInitialization (Variable (TypeInt, vname, 1), InitializationInt i) =
    [StatementExpression (Operation2 ("=", (VariableValue vname), (Number (immediateValue i))))]
emitInitialization (Variable (TypeArray, vname, num), InitializationIntArray is) =
    emitInitializationIntArray 0 vname num is
emitInitialization (_, InitializationNone) = []
emitInitialization _ = emitError "invalid initialization"

emitInitializationIntArray index vname num (i:is) | index < num || num == 0 =
    StatementExpression (Operation2 ("=", ArrayIndex (vname, (Number index)), (Number (immediateValue i)))):
                            (emitInitializationIntArray (index + 1) vname num is)
emitInitializationIntArray index vname num [] = []
emitInitializationIntArray _ _ _ _ = emitError "invalid array initialization"

emitStatements = emitSequence emitStatement

emitSequence ef (t:ts) env context = 
    EmitResult (if ts == [] then EmitValueNone else emitResultValue emitResult,
                (emitResultCode emitResult) ++ emitResultCode nextEmitResult,
                emitResultEnvironment nextEmitResult)
    where
      emitResult = ef t env context
      nextEmitResult = emitSequence ef ts (emitResultEnvironment emitResult) context
emitSequence ef [] env context = EmitResult (EmitValueNone, [], env)

emitStatement tree env context = 
    case tree of
      (If (expression, statements)) -> emitStatementIf (emitExpression expression env (context {emitContextNeedValue = True}))
                                       statements context

      (IfElse (expression, thenStatements, elseStatements)) -> emitStatementIfElse (emitExpression expression env (context {emitContextNeedValue = True}))
                                                               thenStatements elseStatements context

      (While (expression, statements)) -> emitStatementWhile (emitExpression expression env (context {emitContextNeedValue = True}))
                                          statements context

      (InfiniteLoop (statements)) ->
          let
              labelNo = envLabel newEnv
              label1 = emitLabel labelNo
              label2 = emitLabel (labelNo + 1)
              emitResult = emitStatement statements env
                           (context {emitContextBreakLabel = label2})
              newEnv = emitResultEnvironment emitResult
          in
            EmitResult(EmitValueNone,
                       [(Label label1, "infinite loop start")] ++
                       emitResultCode emitResult ++
                       [(FlagResetDontCare, ""),
                        (OperandString1 "JUMP" label1, "infinite loop jump"),
                        (Label label2, "")],
                       incEnvLabel newEnv)

      (Break) -> 
          EmitResult(EmitValueNone,
                     [(FlagResetDontCare, ""),
                      (OperandString1 "JUMP" (emitContextBreakLabel context), "break")],
                     env)

      (StatementExpression expression) -> emitExpression expression env context

      (StatementBlock block) -> emitBlock block env context

      (StatementEmpty) -> EmitResult (EmitValueNone, [], env)

      _ -> emitError ("unkown statement")

emitStatementIf (EmitResult (condVal, condCode, condEnv)) statements context =
    let
        thenResult = emitStatement statements condEnv context
        newEnv = emitResultEnvironment thenResult
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
    in
      EmitResult
      (EmitValueNone,
       condCode ++ emitValueToRegisterA condVal ++
       [(OperandInt1 "AIA" 0xf, "if judege"),
        (OperandInt1 "AIA" 1, ""),
        (OperandString1 "JUMP" label1, "")] ++ 
       emitResultCode thenResult ++
       [(Label label1, "if end")],
       incEnvLabel newEnv)
emitStatementIf (EmitResultAlternative der (EmitResult (EmitValueLabelsFalse condLabels, condCode, condEnv):ers)) statements context =
    let
        thenResult = emitStatement statements condEnv context
        newEnv = emitResultEnvironment thenResult
    in                      
      EmitResult
      (EmitValueNone,
       condCode ++
       emitResultCode thenResult ++
       emitLabels condLabels "if end",
       newEnv)
emitStatementIf (EmitResultAlternative der []) statements context = emitStatementIf der statements context
emitStatementIf (EmitResultAlternative der (er:ers)) statements context = emitStatementIf (EmitResultAlternative der ers) statements context

emitStatementIfElse (EmitResult (condVal, condCode, condEnv)) thenStatements elseStatements context =
    let
        thenResult = emitStatement thenStatements condEnv context
        elseResult = emitStatement elseStatements (emitResultEnvironment thenResult) context
        newEnv = emitResultEnvironment elseResult
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
        label2 = emitLabel (labelNo + 1)
    in
      EmitResult
      (EmitValueNone,
       condCode ++ emitValueToRegisterA condVal ++
       [(OperandInt1 "CIA" 0, "if else judge"),
        (OperandString1 "JUMP" label1, "")] ++ 
       emitResultCode elseResult ++
       [(FlagResetDontCare, ""),
        (OperandString1 "JUMP" label2, "if else jump"),
        (Label label1, "")] ++
       emitResultCode thenResult ++
       [(Label label2, "if else end")],
       (incEnvLabel . incEnvLabel) newEnv)
emitStatementIfElse (EmitResultAlternative der (EmitResult (EmitValueLabelsFalse condLabels, condCode, condEnv):ers)) thenStatements elseStatements context =
    let
        thenResult = emitStatement thenStatements condEnv context
        elseResult = emitStatement elseStatements (emitResultEnvironment thenResult) context
        newEnv = emitResultEnvironment elseResult
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
    in
      EmitResult
      (EmitValueNone,
       condCode ++
       emitResultCode thenResult ++
       [(FlagResetDontCare, ""),
        (OperandString1 "JUMP" label1, "if else jump")] ++
       emitLabels condLabels "else" ++ 
       emitResultCode elseResult ++
       [(Label label1, "if else end")],
       incEnvLabel newEnv)
emitStatementIfElse (EmitResultAlternative der []) thenStatements elseStatements context = emitStatementIfElse der thenStatements elseStatements context
emitStatementIfElse (EmitResultAlternative der (er:ers)) thenStatements elseStatements context = emitStatementIfElse (EmitResultAlternative der ers) thenStatements elseStatements context

emitStatementWhile (EmitResult (condVal, condCode, condEnv)) statements context =
    let
        labelNo = envLabel condEnv
        label1 = emitLabel labelNo
        label2 = emitLabel (labelNo + 1)
        contentResult = emitStatement statements ((incEnvLabel . incEnvLabel) condEnv)
                        (context {emitContextBreakLabel = label2})
        newEnv = emitResultEnvironment contentResult
    in
      EmitResult
      (EmitValueNone,
       [(Label label1, "while start")] ++
       condCode ++ emitValueToRegisterA condVal ++
       [(OperandInt1 "AIA" 0xf, "while judege"),
        (OperandInt1 "AIA" 1, ""),
        (OperandString1 "JUMP" label2, "")] ++
       emitResultCode contentResult ++
       [(FlagResetDontCare, ""),
        (OperandString1 "JUMP" label1, "while loop jump"),
        (Label label2, "while end")],
       newEnv)
emitStatementWhile (EmitResultAlternative der (EmitResult (EmitValueLabelsFalse condLabels, condCode, condEnv):ers)) statements context =
    let
        labelNo = envLabel condEnv
        label1 = emitLabel labelNo
        contentResult = emitStatement statements (incEnvLabel condEnv)
                        (context {emitContextBreakLabel = (head condLabels)})
        newEnv = emitResultEnvironment contentResult
    in
      EmitResult
      (EmitValueNone,
       [(Label label1, "while start")] ++
       condCode ++ 
       emitResultCode contentResult ++
       [(FlagResetDontCare, ""),
        (OperandString1 "JUMP" label1, "while loop jump")] ++
       emitLabels condLabels "while end",
       incEnvLabel newEnv)
emitStatementWhile (EmitResultAlternative der []) statements context = emitStatementWhile der statements context
emitStatementWhile (EmitResultAlternative der (er:ers)) statements context = emitStatementWhile (EmitResultAlternative der ers) statements context

emitExpression tree env context = 
    case tree of
      (Number i) -> EmitResult (EmitValueIntConstant i, [], env)

      (VariableValue vname) | (not.emitContextNeedValue) context -> EmitResult (EmitValueNone, [], env)

      (VariableValue vname) ->
          case contentTable (envTable env) vname of
            (TypeInt, offset) -> EmitResult (EmitValueIntVariable offset vname, [], env)
            (TypeArray, offset) -> EmitResult (EmitValueAddress offset vname, [], env)
            (TypePointer, offset) -> EmitResult (EmitValueAddressInMemory offset vname, [], env)

      PreOperation1 ("+" , expression) -> emitExpression expression env context

      PreOperation1 ("-" , (Number i)) -> EmitResult (EmitValueIntConstant (immediateValue (-i)), [], env)

      PreOperation1 (operator , expression)
          | any (==operator) ["++", "--"] ->
              let
                  emitResult = emitLValue expression env context
                  newEnv = emitResultEnvironment emitResult
              in
                EmitResult(EmitValueIntRegisterA,
                           emitResultCode emitResult ++ 
                           emitAddressToRegisterY (emitResultValue emitResult) ++
                           [(Operand0 "MA", ""),
                            (OperandInt1 "AIA" (if operator == "++" then 1 else 0xf), "Operator '" ++ operator ++ "'"),
                            (Operand0 "AM", "")],
                           newEnv)

      PreOperation1 ("&" , expression) ->
          let
              emitResult = emitLValue expression env context
              newEnv = emitResultEnvironment emitResult
          in
            EmitResult(EmitValueIntRegisterY,
                       emitResultCode emitResult ++
                       (emitAddressToRegisterY (emitResultValue emitResult)),
                       newEnv)
                                          
      PreOperation1 ("!" , expression) -> emitExpressionNot (emitExpression expression env context) context

      PreOperation1 (operator , expression) ->
          let
              emitResult = emitExpression expression env context
              newEnv = emitResultEnvironment emitResult
          in
            case operator of
              _ | (not.emitContextNeedValue) context ->
                  EmitResult(EmitValueNone,
                             emitResultCode emitResult,
                             newEnv)
              "-" ->
                  EmitResult(EmitValueIntRegisterA,
                             emitResultCode emitResult ++
                             (emitValueToRegisterA (emitResultValue emitResult)) ++
                             [(FlagResetRegistersSafe, ""),
                              (OperandString1 "CAL" "CMPL", "Operator '" ++ operator ++ "'"),
                              (OperandInt1 "AIA" 1, "")],
                             newEnv)
              "*" ->
                  EmitResult(EmitValueIntRegisterA,
                             emitResultCode emitResult ++
                             (
                              case emitResultValue emitResult of
                                ev@(EmitValueAddress _ _) -> emitAddressToRegisterY ev
                                ev@(EmitValueAddressRegisterA) -> emitAddressToRegisterY ev
                                ev@(EmitValueAddressRegisterY) -> emitAddressToRegisterY ev
                                ev@_ -> emitValueToRegisterY ev
                             ) ++
                             [(Operand0 "MA", "Operator '" ++ operator ++ "'")],
                             newEnv)
              "~" ->
                  EmitResult(EmitValueIntRegisterA,
                             emitResultCode emitResult ++
                             (emitValueToRegisterA (emitResultValue emitResult)) ++
                             [(FlagResetRegistersSafe, ""),
                              (OperandString1 "CAL" "CMPL", "Operator '" ++ operator ++ "'")],
                             newEnv)

              _ -> emitError ("Operator " ++ operator ++ " not supported")
                  
      PostOperation1 (operator , expression)
          | any (==operator) ["++", "--"] ->
              let
                  emitResult = emitLValue expression env context
                  newEnv = emitResultEnvironment emitResult
              in
                EmitResult(if emitContextNeedValue context then EmitValueIntRegisterA else EmitValueNone,
                           emitResultCode emitResult ++
                           emitAddressToRegisterY (emitResultValue emitResult) ++
                           [(Operand0 "MA", ""),
                            (OperandInt1 "AIA" (if operator == "++" then 1 else 0xf), "Operator '" ++ operator ++ "'"),
                            (Operand0 "AM", "")] ++
                           (if emitContextNeedValue context
                            then [(OperandInt1 "AIA" (if operator == "++" then 0xf else 1), "cancel")]
                            else []
                           ),
                           newEnv)
      PostOperation1 (operator , expression) -> emitError ("Operator " ++ operator ++ " not supported")

      Operation2 ("&&" , expression1, expression2) -> emitExpressionAnd expression1 expression2 env context
      Operation2 ("||" , expression1, expression2) -> emitExpressionOr expression1 expression2 env context
                                                
      Operation2 (">>=" , expression1, expression2) ->
          emitExpression (Operation2 ("=" , expression1, Operation2 (">>", expression1, expression2))) env context

      Operation2 (operator , expression1, expression2) | any (operator==) ["=", "+=", "-="] ->
          let
              emitResult2 = emitExpression expression2 env (context {emitContextNeedValue = True})
              emitResult1 = emitLValue expression1 (incEnvOffset (emitResultEnvironment emitResult2)) (context {emitContextNeedValue = True})
              newEnv = decEnvOffset (emitResultEnvironment emitResult1)
              offset = envOffset newEnv
              labelNo = envLabel newEnv
              label1 = emitLabel labelNo
              label2 = emitLabel (labelNo + 1)
              code1 = emitResultCode emitResult1 ++ emitAddressToRegisterY (emitResultValue emitResult1)
          in
            case any (isBreakRegisterA.fst) code1 of
              False ->
                  EmitResult(EmitValueIntRegisterA,
                             emitResultCode emitResult2 ++
                             emitValueToRegisterA (emitResultValue emitResult2) ++
                             code1 ++
                             (if operator == "+=" || operator == "-="
                              then
                                  [(Operand0 ("M" ++ [head operator]), "Operator '" ++ operator ++ "'"),
                                   (Operand0 "AM", "")]
                              else
                                  [(Operand0 "AM", "Operator '" ++ operator ++ "'")]),
                             newEnv)
              True ->
                  EmitResult(EmitValueIntRegisterA,
                             emitResultCode emitResult2 ++
                             emitValueToRegisterA (emitResultValue emitResult2) ++
                             [(OperandInt1 "TIY" offset, ""),
                              (Operand0 "AM", "")] ++
                             code1 ++
                             [(Operand0 "CH", ""),
                              (OperandInt1 "TIY" offset, ""),
                              (Operand0 "MA", ""),
                              (OperandInt1 "TIY" 0xf, ""),
                              (Operand0 "CH", "copy Br to Yr (too long)"),
                              (Label label2, ""),
                              (OperandInt1 "AIY" 1, ""),
                              (OperandString1 "JUMP" label1, ""),
                              (Operand0 "CH", ""),
                              (OperandInt1 "AIY" 0xf, ""),
                              (Operand0 "CH", ""),
                              (OperandString1 "JUMP" label2, ""),
                              (Label label1, ""),
                              (Operand0 "CH", "")] ++
                              (if operator == "+=" || operator == "-="
                               then
                                   [(Operand0 ("M" ++ [head operator]), "Operator '" ++ operator ++ "'"),
                                    (Operand0 "AM", "")]
                               else
                                   [(Operand0 "AM", "Operator '" ++ operator ++ "'")]),
                              (incEnvLabel.incEnvLabel) newEnv)

      Operation2 (operator , expression1, expression2) ->
          let
              newContext = if operator == "," then context else (context {emitContextNeedValue = True})
              
              emitResult1 = emitExpression expression1 env newContext
              emitResult2 = emitExpression expression2 (incEnvOffset (emitResultEnvironment emitResult1)) newContext
              newEnv = decEnvOffset (emitResultEnvironment emitResult2)
              offset = envOffset newEnv

              reverseEmitResult1 = emitExpression expression2 env newContext
              reverseEmitResult2 = emitExpression expression1 (incEnvOffset (emitResultEnvironment reverseEmitResult1)) newContext
              reverseNewEnv = decEnvOffset (emitResultEnvironment reverseEmitResult2)
              reverseOffset = envOffset reverseNewEnv
          in
            case operator of
              o | any (o==) ["+", "-"] -> let op = if o == "!=" then "-" else o in
                  EmitResult(EmitValueIntRegisterA,
                             emitOperation2ArgumentsCode emitResult1 emitResult2 offset ++
                             [(Operand0 ("M" ++ op), "Operator '" ++ operator ++ "'")],
                             newEnv)

              "," ->
                  EmitResult(emitResultValue emitResult2,
                             emitResultCode emitResult1 ++
                             emitResultCode emitResult2,
                             newEnv)

              ">>" ->
                  case emitResultValue emitResult2 of
                    (EmitValueIntConstant i) ->
                          EmitResult(EmitValueIntRegisterA,
                                     emitResultCode emitResult1 ++
                                     (emitValueToRegisterA (emitResultValue emitResult1)) ++
                                     concat (take i (repeat 
                                     [(FlagResetRegistersSafe, ""),
                                      (OperandString1 "CAL" "SIFT", "Operator '" ++ operator ++ "'")])),
                                     newEnv)
                    _ ->
                        let
                            labelNo = envLabel newEnv
                            label1 = emitLabel labelNo
                            label2 = emitLabel (labelNo + 1)
                        in
                          EmitResult(EmitValueIntRegisterA,
                                     emitResultCode emitResult1 ++
                                     (emitValueToRegisterA (emitResultValue emitResult1)) ++
                                     [(OperandInt1 "TIY" offset, ""),
                                      (Operand0 "AM", "")] ++
                                     emitResultCode emitResult2 ++
                                     (emitValueToRegisterA (emitResultValue emitResult2)) ++
                                     [(FlagResetRegistersSafe, ""),
                                      (OperandString1 "CAL" "CMPL", ""),
                                      (Operand0 "CH", ""),
                                      (OperandInt1 "TIY" offset, ""),
                                      (Operand0 "MA", ""),
                                      (Operand0 "CH", ""),
                                      (Label label2, ""),
                                      (OperandInt1 "AIA" 1, ""),
                                      (OperandString1 "JUMP" label1, ""),
                                      (Operand0 "CH", ""),
                                      (FlagResetRegistersSafe, ""),
                                      (OperandString1 "CAL" "SIFT", "Operator '" ++ operator ++ "'"),
                                      (Operand0 "CH", ""),
                                      (OperandString1 "JUMP" label2, ""),
                                      (Label label1, ""),
                                      (Operand0 "CH", "")],
                                     (incEnvLabel.incEnvLabel) newEnv)

              o | any (o==) [">=", "<=", ">", "<"] ->
                  let
                      emitResultOp1 = (if o == ">=" || o == "<" then emitResult1 else reverseEmitResult1)
                      emitResultOp2 = (if o == ">=" || o == "<" then emitResult2 else reverseEmitResult2)
                      targetNewEnv = (if o == ">=" || o == "<" then newEnv else reverseNewEnv)
                                      
                      labelNo = envLabel targetNewEnv
                      label1 = emitLabel labelNo
                      label2 = emitLabel (labelNo + 1)
                  in
                    EmitResultAlternative
                    (
                     EmitResult
                     (
                      EmitValueIntRegisterA,
                      emitOperation2ArgumentsCode emitResultOp1 emitResultOp2 offset ++
                      [(Operand0 "CH", ""),
                       (if o == "<=" || o == ">="
                        then (OperandInt1 "TIA" 0, "set false")
                        else (OperandInt1 "TIA" 1, "set true")
                       ),
                       (Operand0 "CH", ""),
                       (Operand0 ("M-"), "Operator '" ++ operator ++ "'"),
                       (OperandString1 "JUMP" label1, ""),
                       (if o == "<=" || o == ">="
                        then (OperandInt1 "TIA" 1, "set true")
                        else (OperandInt1 "TIA" 0, "set false")
                       ),
                       (Operand0 "CH", ""),
                       (Label label1, ""),
                       (Operand0 "CH", "")],
                      incEnvLabel targetNewEnv
                     )
                    )

                    (emitResultCodeSizeSort
                     [
                      EmitResult
                      (
                       EmitValueLabelsFalse [label1],
                       emitOperation2ArgumentsCode emitResultOp1 emitResultOp2 offset ++
                       [(Operand0 ("M-"), "Operator '" ++ operator ++ "'")] ++
                       (if o == "<" || o == ">" then [(OperandString1 "JUMP" label2, "")]  else []) ++
                       [(OperandString1 "JUMP" label1, "OOCV")] ++
                       (if o == "<" || o == ">" then [(Label label2, "")]  else []),
                       (incEnvLabel.incEnvLabel) targetNewEnv
                      ),
                      EmitResult
                      (
                       EmitValueLabelsTrue [label1],
                       emitOperation2ArgumentsCode emitResultOp1 emitResultOp2 offset ++
                       [(Operand0 ("M-"), "Operator '" ++ operator ++ "'")] ++
                       (if o == "<=" || o == ">=" then [(OperandString1 "JUMP" label2, "")]  else []) ++
                       [(OperandString1 "JUMP" label1, "OOCV")] ++
                       (if o == "<=" || o == ">=" then [(Label label2, "")]  else []),
                       (incEnvLabel.incEnvLabel) targetNewEnv
                      )
                     ])

              "!=" -> emitExpression (PreOperation1 ("!", Operation2 ("==" , expression1, expression2))) env context

              "==" ->
                  EmitResultAlternative
                  (
                   EmitResult
                   (
                    EmitValueIntRegisterA,
                    emitResultCode emitResult1 ++
                    (emitValueToRegisterA (emitResultValue emitResult1)) ++
                    [(OperandInt1 "TIY" offset, ""),
                     (Operand0 "AM", "")] ++
                    emitResultCode emitResult2 ++
                    (emitValueToRegisterA (emitResultValue emitResult2)) ++
                    [(OperandInt1 "TIY" offset, ""),
                     (FlagResetRegistersSafe, ""),
                     (OperandString1 "CAL" "CHNG", "set false Ar'"),
                     (OperandInt1 "TIA" 0, ""),
                     (OperandInt1 "TIY" offset, ""),
                     (FlagResetRegistersSafe, ""),
                     (OperandString1 "CAL" "CHNG", ""),
                     (Operand0 ("M-"), "Operator '" ++ operator ++ "'"),
                     (OperandInt1 "AIA" 0xf, ""),
                     (OperandString1 "CAL" "CHNG", "")],
                    newEnv
                   )
                  )
                  (
                  let
                      labelNo = envLabel newEnv
                      label1 = emitLabel labelNo
                      label2 = emitLabel (labelNo + 1)
                  in
                    [
                     EmitResult
                     (
                      EmitValueLabelsFalse [label1],
                      case (emitResultValue emitResult1, emitResultValue emitResult2) of
                        ((EmitValueIntConstant constant), _) ->
                            emitResultCode emitResult2 ++
                            (emitValueToRegisterA (emitResultValue emitResult2)) ++
                            [(OperandInt1 "CIA" constant, "Compare constant " ++ show constant ++ " OOCMP=="),
                             (OperandString1 "JUMP" label1, "")]
                        (_, (EmitValueIntConstant constant)) ->
                            emitResultCode emitResult1 ++
                            (emitValueToRegisterA (emitResultValue emitResult1)) ++
                            [(OperandInt1 "CIA" constant, "Compare constant " ++ show constant ++ " OOCMP=="),
                             (OperandString1 "JUMP" label1, "")]
                        _ ->
                            emitOperation2ArgumentsCode emitResult1 emitResult2 offset ++
                            [(Operand0 ("M-"), "Operator '" ++ operator ++ "'"),
                             (OperandInt1 "CIA" 0, ""),
                             (OperandString1 "JUMP" label1, "OOCV")]
                     ,
                      incEnvLabel newEnv
                     ),

                     EmitResult
                     (
                      EmitValueLabelsTrue [label2],
                      case (emitResultValue emitResult1, emitResultValue emitResult2) of
                        ((EmitValueIntConstant constant), _) ->
                            emitResultCode emitResult2 ++
                            (emitValueToRegisterA (emitResultValue emitResult2)) ++
                            [(OperandInt1 "CIA" constant, "Compare constant " ++ show constant ++ " OOCMP=="),
                             (OperandString1 "JUMP" label1, ""),
                             (FlagResetDontCare, ""),
                             (OperandString1 "JUMP" label2, ""),
                             (Label label1, "")]
                        (_, (EmitValueIntConstant constant)) ->
                            emitResultCode emitResult1 ++
                            (emitValueToRegisterA (emitResultValue emitResult1)) ++
                            [(OperandInt1 "CIA" constant, "Compare constant " ++ show constant ++ " OOCMP=="),
                             (OperandString1 "JUMP" label1, ""),
                             (FlagResetDontCare, ""),
                             (OperandString1 "JUMP" label2, ""),
                             (Label label1, "")]
                        _ ->
                            emitOperation2ArgumentsCode emitResult1 emitResult2 offset ++
                            [(Operand0 ("M-"), "Operator '" ++ operator ++ "'"),
                             (OperandInt1 "CIA" 0, ""),
                             (OperandString1 "JUMP" label1, "OOCV"),
                             (FlagResetDontCare, ""),
                             (OperandString1 "JUMP" label2, ""),
                             (Label label1, "")]
                     ,
                      incEnvLabel newEnv
                     )
                    ]
                  )
              _ -> emitError ("operator " ++ operator ++ " not supported")

      (Call (fname, ExpressionEmpty)) ->
              emitExpressionCall0 fname env (context {emitContextNeedValue = True})
      (Call (fname, (Operation2 (",", expression1, expression2)))) ->
              emitExpressionCall2 fname expression1 expression2 env (context {emitContextNeedValue = True})
      (Call (fname, expression)) ->
              emitExpressionCall1 fname expression env (context {emitContextNeedValue = True})

      (ArrayIndex (vname, Number i)) | (not.emitContextNeedValue) context -> 
              EmitResult (EmitValueNone, [], env)

      (ArrayIndex (vname, Number i)) -> 
          case contentTable (envTable env) vname of
            (TypeArray, address) ->
                EmitResult
                (EmitValueIntRegisterA,
                 [(OperandInt1 "TIY" (immediateValue (address + i)), "Variable " ++ vname ++ "[" ++ show i ++ "]"),
                  (Operand0 "MA", "")],
                 env)
            _ -> emitError (vname ++ " is not a array")

      (ArrayIndex (vname, expression)) -> 
          let
              emitResult = emitExpression expression env (context {emitContextNeedValue = True})
              newEnv = emitResultEnvironment emitResult
          in
            case contentTable (envTable env) vname of
              (TypeArray, address) ->
                  EmitResult
                  (EmitValueIntRegisterA,
                   emitResultCode emitResult ++
                   (emitValueToRegisterY (emitResultValue emitResult)) ++
                   [(OperandInt1 "AIY" address, "Offset of " ++ vname ++ "[]"),
                    (Operand0 "MA", "")],
                   newEnv)
              _ -> emitError (vname ++ " is not a array")
                  
      _ -> emitError ("unkown expression")

emitOperation2ArgumentsCode emitResult1 emitResult2 tmpOffset =
    case (emitResultValue emitResult1) of
      (EmitValueIntVariable varOffset varName) ->
          emitResultCode emitResult1 ++
          emitResultCode emitResult2 ++
          (emitValueToRegisterA (emitResultValue emitResult2)) ++
          [(OperandInt1 "TIY" varOffset, "Variable '" ++ varName ++ "'")]
          
      _ ->
          emitResultCode emitResult1 ++
          (emitValueToRegisterA (emitResultValue emitResult1)) ++
          [(OperandInt1 "TIY" tmpOffset, ""),
           (Operand0 "AM", "")] ++
          emitResultCode emitResult2 ++
          (emitValueToRegisterA (emitResultValue emitResult2)) ++
          [(OperandInt1 "TIY" tmpOffset, "")]


emitExpressionNot (EmitResult (expVal, expCode, expEnv)) context =
    let
        labelNo = envLabel expEnv
        label1 = emitLabel labelNo
        label2 = emitLabel (labelNo + 1)
    in        
      (EmitResultAlternative
       (EmitResult
        (EmitValueIntRegisterA,
         expCode ++ emitValueToRegisterA expVal ++
         [(OperandInt1 "CIA" 0, "Operator !"),
          (OperandString1 "JUMP" label1, ""),
          (OperandInt1 "TIA" 1, ""),
          (FlagResetRegistersSafe, ""),
          (OperandString1 "JUMP" label2, ""),
          (Label label1, ""),
          (OperandInt1 "TIA" 0, ""),
          (Label label2, "")],
         (incEnvLabel . incEnvLabel) expEnv)
       )
       [
        EmitResult
        (EmitValueLabelsFalse [label1],
         expCode ++ emitValueToRegisterA expVal ++
         [(OperandInt1 "CIA" 0, "Operator !"),
          (OperandString1 "JUMP" label1, "OOCV")],
         incEnvLabel expEnv),
        
        EmitResult
        (EmitValueLabelsTrue [label1],
         expCode ++ emitValueToRegisterA expVal ++
         [(OperandInt1 "AIA" 0xf, "Operator !"),
          (OperandInt1 "AIA" 1, ""),
          (OperandString1 "JUMP" label1, "OOCV")],
         incEnvLabel expEnv)
       ]
      )

emitExpressionNot (EmitResultAlternative der ers) context =
    (EmitResultAlternative (emitExpressionNot der context) (emitResultCodeSizeSort (map reverseEmitValueLabel ers)))

reverseEmitValueLabel (EmitResult (EmitValueLabelsFalse expLabels, expCode, expEnv)) = (EmitResult (EmitValueLabelsTrue expLabels, expCode, expEnv))
reverseEmitValueLabel (EmitResult (EmitValueLabelsTrue expLabels, expCode, expEnv)) = (EmitResult (EmitValueLabelsFalse expLabels, expCode, expEnv))

emitExpressionAnd expression1 expression2 env context =
    let
        newContext = context {emitContextNeedValue = True}

        emitResult1 = emitResultSelect emitResultIsFalseLabels (emitExpression expression1 env newContext)
        emitResult2 = emitResultSelect emitResultIsFalseLabels (emitExpression expression2 (emitResultEnvironment emitResult1) newContext)
        expVal1 = emitResultValue emitResult1
        expVal2 = emitResultValue emitResult2
        expCode1 = emitResultCode emitResult1
        expCode2 = emitResultCode emitResult2
        newEnv = emitResultEnvironment emitResult2
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
    in
      case (expVal1, expVal2) of
        (EmitValueLabelsFalse expLabels1, EmitValueLabelsFalse expLabels2) ->
            (EmitResultAlternative
             (EmitResult
              (EmitValueIntRegisterA,
               expCode1 ++ 
               expCode2 ++ 
               [(OperandInt1 "TIA" 1, "Operator &&"),
                (FlagResetRegistersSafe, ""),
                (OperandString1 "JUMP" label1, "")] ++
               emitLabels (expLabels1 ++ expLabels2) "" ++
               [(OperandInt1 "TIA" 0, ""),
                (Label label1, "")],
               incEnvLabel newEnv)
             )
             [
              EmitResult
              (EmitValueLabelsFalse (expLabels1 ++ expLabels2),
               expCode1 ++
               expCode2,
               newEnv)
             ]
            )
            
        _ ->
            let
                emitResult1 = emitExpression expression1 env newContext
                emitResult2 = emitExpression expression2 (emitResultEnvironment emitResult1) newContext
                expVal1 = emitResultValue emitResult1
                expVal2 = emitResultValue emitResult2
                expCode1 = emitResultCode emitResult1
                expCode2 = emitResultCode emitResult2
                newEnv = emitResultEnvironment emitResult2
                labelNo = envLabel newEnv
                label1 = emitLabel labelNo
                label2 = emitLabel (labelNo + 1)
            in
              (EmitResultAlternative
               (EmitResult
                (EmitValueIntRegisterA,
                 expCode1 ++ emitValueToRegisterA expVal1 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 1"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, "")] ++
                 expCode2 ++ emitValueToRegisterA expVal2 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 2"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, ""),
                  (OperandInt1 "TIA" 1, ""),
                  (Label label1, "")],
                 incEnvLabel newEnv)
               )
               [
                EmitResult
                (EmitValueLabelsFalse [label1],
                 expCode1 ++ emitValueToRegisterA expVal1 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 1"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, "OOCV")] ++
                 expCode2 ++ emitValueToRegisterA expVal2 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 2"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, "")],
                 incEnvLabel newEnv),
                
                EmitResult
                (EmitValueLabelsTrue [label2],
                 expCode1 ++ emitValueToRegisterA expVal1 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 1"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, "OOCV")] ++
                 expCode2 ++ emitValueToRegisterA expVal2 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 2"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, ""),
                  (OperandString1 "JUMP" label2, ""),
                  (Label label1, "")],
                 (incEnvLabel.incEnvLabel) newEnv)
               ]
              )

emitExpressionOr expression1 expression2 env context =
    let
        newContext = context {emitContextNeedValue = True}

        emitResult1 = emitResultSelect emitResultIsTrueLabels (emitExpression expression1 env newContext)
        emitResult2 = emitResultSelect emitResultIsTrueLabels (emitExpression expression2 (emitResultEnvironment emitResult1) newContext)
        expVal1 = emitResultValue emitResult1
        expVal2 = emitResultValue emitResult2
        expCode1 = emitResultCode emitResult1
        expCode2 = emitResultCode emitResult2
        newEnv = emitResultEnvironment emitResult2
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
        label2 = emitLabel (labelNo + 1)
    in
      case (expVal1, expVal2) of

        (EmitValueLabelsTrue expLabels1, EmitValueLabelsTrue expLabels2) ->
            (EmitResultAlternative
             (EmitResult
              (EmitValueIntRegisterA,
               expCode1 ++ 
               expCode2 ++ 
               [(OperandInt1 "TIA" 0, "Operator ||"),
                (FlagResetRegistersSafe, ""),
                (OperandString1 "JUMP" label1, "")] ++
               emitLabels (expLabels1 ++ expLabels2) "" ++
               [(OperandInt1 "TIA" 1, ""),
                (Label label1, "")],
               incEnvLabel newEnv)
             )
             [
              EmitResult
              (EmitValueLabelsTrue (expLabels1 ++ expLabels2),
               expCode1 ++
               expCode2,
               newEnv),

              EmitResult
              (EmitValueLabelsFalse [label1],
               expCode1 ++
               expCode2 ++
               [(FlagResetRegistersSafe, ""),
                (OperandString1 "JUMP" label1, "Operator ||")] ++
               emitLabels (expLabels1 ++ expLabels2) "",
               incEnvLabel newEnv)
             ]
            )

        _ ->
            let
                emitResult1 = emitExpression expression1 env newContext
                emitResult2 = emitExpression expression2 (emitResultEnvironment emitResult1) newContext
                expVal1 = emitResultValue emitResult1
                expVal2 = emitResultValue emitResult2
                expCode1 = emitResultCode emitResult1
                expCode2 = emitResultCode emitResult2
                newEnv = emitResultEnvironment emitResult2
                labelNo = envLabel newEnv
                label1 = emitLabel labelNo
                label2 = emitLabel (labelNo + 1)
            in
              (EmitResultAlternative
               (EmitResult
                (EmitValueIntRegisterA,
                 expCode1 ++ emitValueToRegisterA expVal1 ++
                 [(OperandInt1 "CIA" 0, "Operator || 1"),
                  (OperandString1 "JUMP" label1, "")] ++
                 expCode2 ++ emitValueToRegisterA expVal2 ++
                 [(OperandInt1 "CIA" 0, "Operator || 2"),
                  (OperandString1 "JUMP" label1, ""),
                  (OperandInt1 "TIA" 0, ""),
                  (Label label1, "")],
                 incEnvLabel newEnv)
               )
               [
                EmitResult
                (EmitValueLabelsTrue [label1],
                 expCode1 ++ emitValueToRegisterA expVal1 ++
                 [(OperandInt1 "CIA" 0, "Operator || 1"),
                  (OperandString1 "JUMP" label1, "OOCV")] ++
                 expCode2 ++ emitValueToRegisterA expVal2 ++
                 [(OperandInt1 "CIA" 0, "Operator || 2"),
                  (OperandString1 "JUMP" label1, "")],
                 incEnvLabel newEnv),

                EmitResult
                (EmitValueLabelsFalse [label2],
                 expCode1 ++ emitValueToRegisterA expVal1 ++
                 [(OperandInt1 "CIA" 0, "Operator || 1"),
                  (OperandString1 "JUMP" label1, "OOCV")] ++
                 expCode2 ++ emitValueToRegisterA expVal2 ++
                 [(OperandInt1 "CIA" 0, "Operator || 2"),
                  (OperandString1 "JUMP" label1, ""),
                  (FlagResetDontCare, ""),
                  (OperandString1 "JUMP" label2, ""),
                  (Label label1, "")],
                 (incEnvLabel.incEnvLabel) newEnv)
               ]
              )

emitExpressionCall0 fname env context = 
    case fname of
      "exit" ->
          let
              labelNo = envLabel env
              label1 = emitLabel labelNo
          in
            EmitResult(EmitValueVoid,
                       [(FlagResetDontCare, ""),
                        (Label label1, ""),
                        (OperandString1 "JUMP" label1, "exit loop")],
                       incEnvLabel env)
      "inputKeyBlock" ->
          let
              labelNo = envLabel env
              label1 = emitLabel labelNo
          in
            EmitResult(EmitValueIntRegisterA,
                       [(Label label1, ""),
                        (Operand0 "KA", ""),
                        (OperandString1 "JUMP" label1, "")],
                       incEnvLabel env)
          
      _ ->
          EmitResult (EmitValueVoid, [(FlagResetDontCare, "")] ++ emitFunctionCall0 fname, env)

emitFunctionCall0 fn@"resetNumberLed" = [(OperandString1 "CAL" "RSTO", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepEndSound" = [(OperandString1 "CAL" "ENDS", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepErrorSound" = [(OperandString1 "CAL" "ERRS", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepShortSound" = [(OperandString1 "CAL" "SHTS", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepLongSound" = [(OperandString1 "CAL" "LONS", "call " ++ fn ++ "()")]
emitFunctionCall0 fname = emitError ("unkown function " ++ fname ++ "(void)")

emitExpressionCall1 fname expression env context = 
    case fname of

      "inputKey" | expression == (Number 0) ->
          let
              labelNo = envLabel env
              label1 = emitLabel labelNo
              label2 = emitLabel (labelNo + 1)
          in
            EmitResultAlternative
            (
             EmitResult
             (EmitValueIntRegisterA,
              [(OperandInt1 "TIA" 0, "inputKey()"),
               (Operand0 "CH" , ""),
               (Operand0 "KA", ""),
               (OperandString1 "JUMP" label1, ""),
               (OperandInt1 "TIA" 1, ""),
               (Operand0 "CH" , ""),
               (Label label1, ""),
               (Operand0 "CH" , "")],
              incEnvLabel env
             )
            )
            [
             EmitResult
             (
              EmitValueLabelsFalse [label1],
              [(Operand0 "KA", "inputKey()"),
               (OperandString1 "JUMP" label1, "OOCV")],
              incEnvLabel env
             ),
             EmitResult
             (
              EmitValueLabelsTrue [label2],
              [(Operand0 "KA", "inputKey()"),
               (OperandString1 "JUMP" label1, "OOCV"),
               (FlagResetDontCare, ""),
               (OperandString1 "JUMP" label2, ""),
               (Label label1, "")],
              (incEnvLabel.incEnvLabel) env
             )
            ]

      "inputKey" ->
          let
              emitResult = emitExpression expression env (context {emitContextNeedValue = True})
              newEnv = emitResultEnvironment emitResult
              addressCalcCode = emitResultCode emitResult
              labelNo = envLabel newEnv
              label1 = emitLabel labelNo
              label2 = emitLabel (labelNo + 1)
          in
            case addressCalcCode of
              [(OperandInt1 "TIY" _, _)] ->
                  EmitResultAlternative
                  (
                   EmitResult
                   (EmitValueIntRegisterA,
                    [(OperandInt1 "TIA" 0, "inputKey()"),
                     (Operand0 "CH" , ""),
                     (Operand0 "KA", ""),
                     (OperandString1 "JUMP" label1, "")] ++
                    addressCalcCode ++
                    [(Operand0 "AM", ""),
                     (OperandInt1 "TIA" 1, ""),
                     (Operand0 "CH" , ""),
                     (Label label1, ""),
                     (Operand0 "CH" , "")],
                    incEnvLabel newEnv
                   )
                  )
                  [
                   EmitResult
                   (
                    EmitValueLabelsFalse [label1],
                    [(Operand0 "KA", ""),
                     (OperandString1 "JUMP" label1, "OOCV")] ++
                    addressCalcCode ++
                    [(Operand0 "AM", "")],
                    incEnvLabel newEnv
                   ),
                   EmitResult
                   (
                    EmitValueLabelsTrue [label2],
                    [(Operand0 "KA", ""),
                     (OperandString1 "JUMP" label1, "OOCV")] ++
                    addressCalcCode ++
                    [(Operand0 "AM", ""),
                     (FlagResetDontCare, ""),
                     (OperandString1 "JUMP" label2, ""),
                     (Label label1, "")],
                    (incEnvLabel.incEnvLabel) newEnv
                   )
                  ]

              _ -> emitError ("can't emit expression in inputKey(..), code size too large")
      _ ->
          let
              emitResult = emitExpression expression env (context {emitContextNeedValue = True})
          in
            EmitResult(EmitValueVoid,
                       emitResultCode emitResult ++
                       emitValueToRegisterA (emitResultValue emitResult) ++
                       emitFunctionCall1 fname,
                       emitResultEnvironment emitResult)

emitFunctionCall1 fn@"setNumberLed" = [(Operand0 "AO", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"setLed" = [(Operand0 "CY", ""), (OperandString1 "CAL" "SETR", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"resetLed" = [(Operand0 "CY", ""), (OperandString1 "CAL" "RSTR", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"beep" = [(FlagResetRegistersSafe, "")] ++ [(OperandString1 "CAL" "SUND", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"sleep" = [(FlagResetRegistersSafe, "")] ++ [(OperandString1 "CAL" "TIMR", "call " ++ fn ++ "()")]
emitFunctionCall1 fname = emitError ("unkown function " ++ fname ++ "(int)")

emitExpressionCall2 fname expression1 expression2 env context = 
    case fname of
      "setLedBinary" ->
          let
              emitResult1 = emitExpression expression1 env (context {emitContextNeedValue = True})
              emitResult2 = emitExpression expression2 (emitResultEnvironment emitResult1) (context {emitContextNeedValue = True})
              newEnv = emitResultEnvironment emitResult2
          in -- ignore "setLedBinary((setLedBinary(i, j),j), i);" pattern
            EmitResult(EmitValueVoid,
                       emitResultCode emitResult1 ++
                       case emitResultValue emitResult1 of
                         (EmitValueIntVariable 0xf _) -> []
                         _ -> (emitValueToRegisterA (emitResultValue emitResult1)) ++
                              [(OperandInt1 "TIY" 0xf, ""),
                               (Operand0 "AM", "")]
                       ++
                       emitResultCode emitResult2 ++
                       case emitResultValue emitResult2 of
                         (EmitValueIntVariable 0xe _) -> []
                         _ -> (emitValueToRegisterA (emitResultValue emitResult2)) ++
                              [(OperandInt1 "TIY" 0xe, ""),
                               (Operand0 "AM", "")]
                       ++
                       [(FlagResetDontCare, ""),         
                        (OperandString1 "CAL" "DSPR", "call " ++ fname ++ "()")],
                       newEnv)
      _ -> emitError ("unkown function " ++ fname ++ "(int, int)")

emitLValue (PreOperation1 ("*", expression)) env context =
    let
        emitResult = emitExpression expression env (context {emitContextNeedValue = True})
        newEnv = emitResultEnvironment emitResult
    in
      EmitResult
      (EmitValueAddressRegisterY,
       emitResultCode emitResult ++
       (
        case emitResultValue emitResult of
          ev@(EmitValueAddress _ _) -> emitAddressToRegisterY ev
          ev@(EmitValueAddressRegisterA) -> emitAddressToRegisterY ev
          ev@(EmitValueAddressRegisterY) -> emitAddressToRegisterY ev
          ev@_ -> emitValueToRegisterY ev
       ),
       newEnv)
    
emitLValue (VariableValue vname) env context =
    case contentTable (envTable env) vname of
      (TypeInt, offset) -> EmitResult (EmitValueAddress offset vname, [], env)
      (TypePointer, address) -> EmitResult (EmitValueAddress address vname, [], env)
      _ -> emitError ("invalid lvalue " ++ show vname)

emitLValue (ArrayIndex (vname, Number i)) env context =
    case contentTable (envTable env) vname of
      (TypeArray, address) ->
          EmitResult (EmitValueAddress (immediateValue (address + i)) vname, [], env)
      _ -> emitError (vname ++ " is not a array")

emitLValue (ArrayIndex (vname,expression)) env context =
    let
        emitResult = emitExpression expression env (context {emitContextNeedValue = True})
        newEnv = emitResultEnvironment emitResult
    in
      case contentTable (envTable env) vname of
        (TypeArray, address) ->
            EmitResult
            (EmitValueAddressRegisterY,
             emitResultCode emitResult ++
             (emitValueToRegisterY (emitResultValue emitResult)) ++
             [(OperandInt1 "AIY" address, "Offset of " ++ vname ++ "[]")],
             newEnv)
        _ -> emitError (vname ++ " is not a array")

emitLValue t _ _ = emitError ("invalid lvalue " ++ show t)

emitValueToRegisterA EmitValueVoid = emitError ("void value not ignored as it ought to be")
emitValueToRegisterA EmitValueIntRegisterA = []
emitValueToRegisterA EmitValueIntRegisterY = [(Operand0 "CY", "")]
emitValueToRegisterA (EmitValueIntTemporaryMemory i) =
    if i < 0
    then emitError "memory over"
    else [(OperandInt1 "TIY" i, "Temporary offset " ++ show i) , (Operand0 "MA", "")]
emitValueToRegisterA (EmitValueIntVariable i s) =
    [(OperandInt1 "TIY" i, "Variable '" ++ s ++ "' (offset " ++ show i ++ ")") , (Operand0 "MA", "")]
emitValueToRegisterA (EmitValueIntConstant i) = 
    [(OperandInt1 "TIA" i, "Constant " ++ show i)]
emitValueToRegisterA EmitValueAddressRegisterA = []
emitValueToRegisterA EmitValueAddressRegisterY = [(Operand0 "CY", "")]
emitValueToRegisterA (EmitValueAddress i s) =
    [(OperandInt1 "TIA" i, "Address of '" ++ s ++ "' (offset " ++ show i ++ ")")]
emitValueToRegisterA (EmitValueAddressInMemory i s) =
    [(OperandInt1 "TIY" i, "Pointer '" ++ s ++ "' (offset " ++ show i ++ ")"),
     (Operand0 "MA", "")]

emitValueToRegisterA t = emitError ("invalid value " ++ show t)

emitValueToRegisterY (EmitValueIntConstant i) = 
    [(OperandInt1 "TIY" i, "Constant " ++ show i)]
emitValueToRegisterY EmitValueIntRegisterY = []
emitValueToRegisterY EmitValueAddressRegisterY = []
emitValueToRegisterY (EmitValueAddress i s) =
    [(OperandInt1 "TIY" i, "Address of '" ++ s ++ "' (offset " ++ show i ++ ")")]
emitValueToRegisterY x = (emitValueToRegisterA x) ++ [(Operand0 "CY", "")]

emitAddressToRegisterA = emitValueToRegisterA
emitAddressToRegisterY = emitValueToRegisterY

emitLabel = ("LABEL" ++) . show

emitLabels (l:ls) comment = (Label l, comment) : emitLabels (ls) ""
emitLabels [] _ = []

emitResultCodeSizeSort = sortBy emitResultCodeSizeCompare
emitResultCodeSizeCompare (EmitResult (_, a, _)) (EmitResult (_, b, _)) =
    compare (sum (map (instructionSize.fst) a)) (sum (map (instructionSize.fst) b))

immediateValue i | i < 0 = mod (0xf + 1 + i) 0x10
immediateValue i = mod i 0x10

getSetLedBinaryArgumentsProgram (Program globals) =
    nub $ concat $ map getSetLedBinaryArgumentsGlobal globals
getSetLedBinaryArgumentsGlobal (GlobalDeclaration _) = []
getSetLedBinaryArgumentsGlobal (GlobalFunction function) = getSetLedBinaryArgumentsFunction function
getSetLedBinaryArgumentsFunction (Function (fname, ftype, args, block)) = 
    getSetLedBinaryArgumentsBlock block
getSetLedBinaryArgumentsBlock (Block (decs, statemants)) = concat $ map getSetLedBinaryArgumentsStatemant statemants

getSetLedBinaryArgumentsStatemant (If (expression, statement)) =
    getSetLedBinaryArgumentsExpression expression ++ getSetLedBinaryArgumentsStatemant statement
getSetLedBinaryArgumentsStatemant (IfElse (expression, thenStatement, elseStatement)) =
    getSetLedBinaryArgumentsExpression expression ++ getSetLedBinaryArgumentsStatemant thenStatement ++
    getSetLedBinaryArgumentsStatemant elseStatement
getSetLedBinaryArgumentsStatemant (While (expression, statement)) =
    getSetLedBinaryArgumentsExpression expression ++ getSetLedBinaryArgumentsStatemant statement
getSetLedBinaryArgumentsStatemant (InfiniteLoop statement) =
    getSetLedBinaryArgumentsStatemant statement
getSetLedBinaryArgumentsStatemant (StatementBlock block) =
    getSetLedBinaryArgumentsBlock block
getSetLedBinaryArgumentsStatemant (StatementExpression expression) =
    getSetLedBinaryArgumentsExpression expression
getSetLedBinaryArgumentsStatemant statement = []

getSetLedBinaryArgumentsExpression (PreOperation1 (op, expression)) = getSetLedBinaryArgumentsExpression expression
getSetLedBinaryArgumentsExpression (PostOperation1 (op, expression)) = getSetLedBinaryArgumentsExpression expression
getSetLedBinaryArgumentsExpression (Operation2 (op, expression1, expression2)) =
    getSetLedBinaryArgumentsExpression expression1 ++ getSetLedBinaryArgumentsExpression expression2
getSetLedBinaryArgumentsExpression (Call ("setLedBinary", expression)) = [expression]
getSetLedBinaryArgumentsExpression (Call (fname, expression)) = getSetLedBinaryArgumentsExpression expression
getSetLedBinaryArgumentsExpression (ArrayIndex (vname, expression)) = getSetLedBinaryArgumentsExpression expression
getSetLedBinaryArgumentsExpression expression = []
