module Lexer where

import Char
import List

data Token =
    TokenIdentifier String |
    TokenSymbol String |
    TokenNumber Int |
    TokenNewLine
    deriving (Show, Eq)

lexer = removeNewLine . removeComment . firstLex

lexError = error . ("lexer error: " ++)

removeNewLine (TokenNewLine:tokens) = removeNewLine tokens
removeNewLine (token:tokens) = token : removeNewLine tokens
removeNewLine [] = []

removeComment = removeLineComment . removeBlockComment

removeLineComment = removeTokensBetween (==(TokenSymbol "//")) (==TokenNewLine) 0
removeBlockComment = removeTokensBetween (==(TokenSymbol "/*")) (==(TokenSymbol "*/")) 0

removeTokensBetween startTokenPredicate endTokenPredicate n (token:tokens)
    | startTokenPredicate token && n == 0 = removeTokensBetween startTokenPredicate endTokenPredicate (n + 1) tokens
removeTokensBetween startTokenPredicate endTokenPredicate n (token:tokens)
    | endTokenPredicate token && n > 0 = removeTokensBetween startTokenPredicate endTokenPredicate (n - 1) tokens
removeTokensBetween startTokenPredicate endTokenPredicate n (token:tokens)
    | n > 0 = removeTokensBetween startTokenPredicate endTokenPredicate n tokens
removeTokensBetween startTokenPredicate endTokenPredicate n (token:tokens) =
    token : removeTokensBetween startTokenPredicate endTokenPredicate n tokens
removeTokensBetween startTokenPredicate endTokenPredicate 0 [] = []
removeTokensBetween startTokenPredicate endTokenPredicate n [] = lexError "unterminated comment"

firstLex [] = []
firstLex s =
    if fst number /= "" then
        TokenNumber (readNumber (fst number)) : firstLex (snd number)
    else if fst alphas /= "" then
             TokenIdentifier (fst alphas) : firstLex (snd alphas)
         else if fst newline /= "" then
                  TokenNewLine : firstLex (snd newline)
              else if fst symbols /= "" then
                       TokenSymbol (fst symbols) : firstLex (snd symbols)
                   else if fst symbol /= "" then
                            TokenSymbol (fst symbol) : firstLex (snd symbol)
                        else
                            firstLex (tail s)
    where
      alwaysFalse x = False

      isIdSymbols x = any (==x) ['_']

      isIdFirstChar x = isAlpha x || isIdSymbols x
      isIdChar x = isAlphaNum x || isIdSymbols x
                      
      alphas = tokenerWord isIdFirstChar isIdChar s ""
      number = tokenerWord isDigit isAlphaNum s ""
      newline = tokenerWord isNewLine alwaysFalse s ""
          where
            isNewLine x = any (x==) "\r\n"
      symbol = tokenerWord (\x -> (not.isIdChar) x && (not.isSpace) x) alwaysFalse s ""
      symbols = findSymbols multiSymbols s
      multiSymbols = 
          ["++", "--", "||", "&&",
           "==", "!=", "<=", ">=", "+=", "-=",
           ">>=", ">>",
           "//", "/*", "*/"]
      findSymbols (sym:syms) cs | length cs >= length sym && sym == take (length sym) cs = (sym, (tails cs)!!(length sym))
      findSymbols (sym:syms) cs = findSymbols (syms) cs
      findSymbols [] cs = ("", cs)

tokenerWord fp sp (c:cs) "" | fp c = tokenerWord fp sp cs [c]
tokenerWord fp sp (c:cs) "" = ("", cs)
tokenerWord fp sp (c:cs) s | sp c = tokenerWord fp sp cs (s ++ [c])
tokenerWord fp sp cs s = (s, cs)

readNumber ('0':'x':s) = readHexString s
readNumber ('0':'b':s) = readBinaryString s
readNumber ('0':s) = readOctString s
readNumber s = read s

-- char code free

hexChar i | i < 0 = error ("invalid integer " ++ show i ++ " or memory over")
hexChar i | i < 16  = ["0123456789ABCDEF" !! i]
hexChar i = error ("invalid integer " ++ show i)

hexString n 0 | n <= 0 = ""
hexString n 0 = "0" ++ hexString (n-1) 0
hexString n i = hexString (n - 1) (div i 16) ++ hexChar (mod i 16)

readNumberString = readNumberString' 0
readNumberString' n cf base (c:cs) = readNumberString' (n * base + cf [c]) cf base cs
readNumberString' n cf base [] = n::Int

readHexString = readNumberString readHexChar 16
readHexChar "0" = 0
readHexChar "1" = 1
readHexChar "2" = 2
readHexChar "3" = 3
readHexChar "4" = 4
readHexChar "5" = 5
readHexChar "6" = 6
readHexChar "7" = 7
readHexChar "8" = 8
readHexChar "9" = 9
readHexChar "A" = 10
readHexChar "B" = 11
readHexChar "C" = 12
readHexChar "D" = 13
readHexChar "E" = 14
readHexChar "F" = 15
readHexChar "a" = 10
readHexChar "b" = 11
readHexChar "c" = 12
readHexChar "d" = 13
readHexChar "e" = 14
readHexChar "f" = 15
readHexChar s = lexError ("invalid hexadecimal constant" ++ s)

readOctString = readNumberString readOctChar 8
readOctChar "0" = 0
readOctChar "1" = 1
readOctChar "2" = 2
readOctChar "3" = 3
readOctChar "4" = 4
readOctChar "5" = 5
readOctChar "6" = 6
readOctChar "7" = 7
readOctChar s = lexError ("invalid Octal constant" ++ s)

readBinaryString = readNumberString readBinaryChar 2
readBinaryChar "0" = 0
readBinaryChar "1" = 1
readBinaryChar s = lexError ("invalid binary constant" ++ s)
