module Optimizer where

import Parser
import Instruction
import Emitter
import Coder

import Debug.Trace
dump x = trace (show x) x

treeOptimize = optimizeRepeat optimizeProgram

optimizeProgram (Program globals) =
    (Program (map optimizeGlobal globals))
optimizeGlobal (GlobalFunction function) = (GlobalFunction (optimizeFunction function))
optimizeGlobal x = x

optimizeFunction (Function (fname, ftype, args, block)) = 
    (Function (fname, ftype, args, optimizeBlock block))

optimizeBlock (Block (decs, statemants)) = Block (decs, map optimizeStatemant statemants)

optimizeStatemant (If (expression, statement)) =
    If (optimizeExpression expression, optimizeStatemant statement)
optimizeStatemant (IfElse (expression, thenStatement, elseStatement)) =
    IfElse (optimizeExpression expression, optimizeStatemant thenStatement, optimizeStatemant elseStatement)
optimizeStatemant (While ((Number 0), statement)) =
    StatementEmpty
optimizeStatemant (While ((Number i), statement)) =
    optimizeStatemant (InfiniteLoop statement)
optimizeStatemant (While (expression, statement)) =
    While (optimizeExpression expression, optimizeStatemant statement)
optimizeStatemant (InfiniteLoop statement) =
    InfiniteLoop (optimizeStatemant statement)
optimizeStatemant (StatementBlock block) =
    StatementBlock (optimizeBlock block)
optimizeStatemant (StatementExpression expression) =
    StatementExpression (optimizeExpression expression)
optimizeStatemant statement = statement

optimizeExpression (Operation2 ("=", (VariableValue vname1), (Operation2 (op, (VariableValue vname2), expression))))
    | vname1 == vname2 && any (==op) ["+", "-"] = 
    Operation2 (op ++ "=", (VariableValue vname1), optimizeExpression expression)
optimizeExpression (Operation2 ("=", (VariableValue vname1), (Operation2 (op, expression, (VariableValue vname2)))))
    | vname1 == vname2 && any (==op) ["+"] = 
    Operation2 (op ++ "=", (VariableValue vname1), optimizeExpression expression)
optimizeExpression (Operation2 ("+", Number i1, Number i2)) = Number (immediateValue (i1 + i2))
optimizeExpression (Operation2 ("-", Number i1, Number i2)) = Number (immediateValue (i1 - i2))
optimizeExpression (Operation2 (op, Number i1, expression))
    | any (==op) ["+", "==", "!="] = Operation2 (op, expression, Number i1)

optimizeExpression (PreOperation1 ("-", Number i)) = Number (immediateValue (-i))
optimizeExpression (PreOperation1 (op, expression)) = PreOperation1 (op, optimizeExpression expression)
optimizeExpression (PostOperation1 (op, expression)) = PostOperation1 (op, optimizeExpression expression)
optimizeExpression (Operation2 (op, expression1, expression2)) = Operation2 (op, optimizeExpression expression1, optimizeExpression expression2)
optimizeExpression (Call (fname, expression)) = Call (fname, optimizeExpression expression)
optimizeExpression (ArrayIndex (vname, expression)) = ArrayIndex (vname, optimizeExpression expression)
optimizeExpression expression = expression

macheCodesOptimize = resolvFlagReset . (optimizeRepeat optimizeBlockLevel)
macheCodesRequiredOptimize = resolvFlagReset

optimizeBlockLevel = concat . optimizeBlocks . splitBlocks

optimizeBlocks =
    map (optimizeAMMABlock.optimizeRegisterAValueBlock.optimizeRegisterYValueBlock)

optimizeRegisterYValueBlock = optimizeRegisterValueBlock isBreakRegisterY "TIY"
optimizeRegisterAValueBlock = optimizeRegisterValueBlock isBreakRegisterA "TIA"

optimizeRegisterValueBlock p ins = optimizeRegisterValueBlock' p ins [] []

optimizeRegisterValueBlock' p ins _ result [] = result
optimizeRegisterValueBlock' p ins [c] result (i@(OperandInt1 cins v, comment):is)
    | c == v && ins == cins =
        if or (map (p.fst) result) then
           result ++ [i] ++  optimizeRegisterValueBlock' p ins [v] []  is
        else
           result ++ optimizeRegisterValueBlock' p ins [v] [] (appendComment (" OO" ++ ins) is)
optimizeRegisterValueBlock' p ins cc result (i@(OperandInt1 cins v, comment):is)
    | ins == cins =
    result ++ [i] ++ optimizeRegisterValueBlock' p ins [v] [] is
optimizeRegisterValueBlock' p ins cc result (i:is) =
    optimizeRegisterValueBlock' p ins cc (result ++ [i]) is

optimizeAMMABlock [] = []
optimizeAMMABlock (i@(Operand0 "AM", _):is) =
    if snd b == []
    then
        i:is
    else
        case snd b of
          ((Operand0 "MA", _):sis) -> i:(fst b) ++ (optimizeAMMABlock (appendComment " OOMA" sis))
          _ -> i:optimizeAMMABlock is
    where
      b = break (\(x,y) -> isBreakRegisterA x || isBreakRegisterY x || isBreakMemory x) is
optimizeAMMABlock (i:is) = i : optimizeAMMABlock is

splitBlocks [] = []
splitBlocks is =
    case s of
      [] -> [f]
      _ -> f : [head s] : splitBlocks (tail s)
    where
      b = break (\x -> isLabel (fst x)) is
      f = fst b
      s = snd b

showBlock :: [(Instruction, String)] -> String
showBlock = unlines.(map codingMnemonic)
showBlocks = unlines . map (++(take 60 (repeat '-'))) . (map showBlock)

resolvFlagReset = resolvFlagReset' [] []
resolvFlagReset' [c] ps [] = c: ps
resolvFlagReset' [] ps [] = ps
resolvFlagReset' [(ci, ccomment)] ps ((ri, rcomment):is)
    | isResetFlag ci && (ri == FlagResetDontCare || ri == FlagResetRegistersSafe) = 
        resolvFlagReset' [(ci, ccomment ++ " OORF")] ps is
resolvFlagReset' cc ps ((FlagResetDontCare, rcomment):is) =
    cc ++ ps ++ [(Operand0 "CH", "reset flag")] ++ resolvFlagReset' [] [] is
resolvFlagReset' cc ps ((FlagResetRegistersSafe, rcomment):is) =
    cc ++ ps ++ [(Operand0 "CH", "reset flag safe"), (Operand0 "CH", "")] ++ resolvFlagReset' [] [] is
resolvFlagReset' cc ps (i:is) | (not.isInstruction) (fst i) =
    resolvFlagReset' cc (ps++[i]) is
resolvFlagReset' cc ps (rr:is) =
    cc ++ ps ++ resolvFlagReset' [rr] [] is

optimizeRepeat f x =
    if x == n then x else optimizeRepeat f n
    where
      n = f x

appendComment apc [] = []
appendComment apc ((ins, comment):is) = (ins, comment ++ apc):is
